from sklearn.base import BaseEstimator

class Surrogate(BaseEstimator):
    '''
    Surrogate base class.
    '''
    def __init__(self):
        super(Surrogate, self).__init__()
    
    def fit(self, X, y):
        '''
        Fit the surrogate model on a training set (X, y).

        Parameters
        ----------
        X : np.array of shape (n_samples * func.n_obs, func.dim)
            Training input points.

        y : np.array of shape (n_samples * func.n_obs, func.dim)
            Target values of the black-box function.
        '''
        pass
    
    def predict(self, X):
        '''
        Preedict black-box function values.

        Parameters
        ----------
        X : np.array of shape (n_samples * func.n_obs, func.dim)
            Candidate points.
        
        Returns
        -------
        y : np.array of shape (n_samples * func.n_obs, 1)
            Predicted values of the black-box function
        '''
        pass

class GenSurrogate(Surrogate):
    '''
    Generative surrogate base class.
    '''
    def __init__(self):
        super(GenSurrogate, self).__init__()
        pass
    
    def sample(self, X):
        '''
        Generate black-box response.

        Parameters
        ----------
        X : np.array of shape (n_samples * func.n_obs, func.dim)
            Candidate points.
        
        Returns
        -------
        y : np.array of shape (n_sample * func.n_obs, 1)
            Generated response.
        '''
        pass



import gc
import numpy as np
from .surrogate import SurrogateOptimiser


class BarycentreSurrogateOptimiser(SurrogateOptimiser):
    def __init__(self, func, surr, acqf, **kwargs):
        super(BarycentreSurrogateOptimiser, self).__init__(func, surr, acqf, **kwargs)
        self.acqf.n_epochs = self.surr.n_epochs
        self.save_iter = 0
        self.save_coef = 0.0
    
    def get_lip(self, X, y):
        idx = np.array(np.meshgrid(np.arange(X.shape[0]), np.arange(X.shape[0]))).T.reshape(-1, 2)
        idx = idx[idx[:, 0] != idx[:, 1]]
        L = np.max(np.linalg.norm(y[idx[:, 0]] - y[idx[:, 1]], axis=-1) / np.linalg.norm(X[idx[:, 0]] - X[idx[:, 1]], axis=-1))
        return L

    def predict(self, X, y):

        if type(self.acqf.wp) == str:
            self.surr.save_epoch = min(int(self.save_iter * self.save_coef), self.surr.n_epochs - 2)
            self.save_iter += 1
            self.save_coef += 0.01
        else:
            self.surr.save_epoch = int(self.surr.n_epochs * self.acqf.wp)

        self.surr.fit(X, y)
        self.acqf.surr = [self.surr]
        self.acqf.L = self.get_lip(X, y)
        
        x0 = None
        if self.num_opt_start == 'fmin':
            x0 = X[np.argmin(y)]
        
        next_x = self.numerical_search(x0=x0)

        del self.acqf.surr
        gc.collect()
        
        if next_x in X:
            next_x += np.random.normal(0, self.jitter, 1)
        
        return np.array([next_x])


class EnsembleBarycentreSurrogateOptimiser(SurrogateOptimiser):
    def __init__(self, func, surr, acqf, **kwargs):
        super(EnsembleBarycentreSurrogateOptimiser, self).__init__(func, surr, acqf, **kwargs)
        
        for surr in self.surr:
            surr.verbose   = self.verbose
        self.acqf.verbose   = self.verbose
    
    def get_lip(self, X, y):
        idx = np.array(np.meshgrid(np.arange(X.shape[0]), np.arange(X.shape[0]))).T.reshape(-1, 2)
        idx = idx[idx[:, 0] != idx[:, 1]]
        L = np.max(np.linalg.norm(y[idx[:, 0]] - y[idx[:, 1]], axis=-1) / np.linalg.norm(X[idx[:, 0]] - X[idx[:, 1]], axis=-1))
        return L

    def predict(self, X, y):
        
        surrs = []
        
        for surr in self.surr:
            surr.fit(X, y)
            surrs.append(surr)
        
        self.acqf.surr = surrs
        self.acqf.L = self.get_lip(X, y)
        
        x0 = None
        if self.num_opt_start == 'fmin':
            x0 = X[np.argmin(y)]
        
        next_x = self.numerical_search(x0=x0)
        
        del self.acqf.surr
        gc.collect()

        return np.array([next_x])



from .base import Surrogate

import torch
import numpy as np
from tqdm import tqdm
import torch.nn as nn
import torchbnn as bnn
from torch.utils.data import TensorDataset, DataLoader

# cuda = True if torch.cuda.is_available() else False
Tensor = torch.FloatTensor #torch.cuda.FloatTensor if cuda else


class BayesianRegressor(nn.Module):
    def __init__(self, n_inputs=2, n_outputs=1, hidden_size=64):
        super(BayesianRegressor, self).__init__()

        self.model = nn.Sequential(
            bnn.BayesLinear(prior_mu=0, prior_sigma=0.1, in_features=n_inputs, out_features=hidden_size),
            nn.Tanh(),
            bnn.BayesLinear(prior_mu=0, prior_sigma=0.1, in_features=hidden_size, out_features=n_outputs)
        )
    
    def forward(self, x):
        x = self.model(x)
        return x

class BNN(Surrogate):# TODO: add cuda
    def __init__(self, **kwargs):
        super(BNN, self).__init__()

        self.name        = 'BNN'
        self.model       = kwargs['model'] if 'model' in kwargs else None
        self.hidden_size = kwargs['hidden_size'] if 'hidden_size' in kwargs else 64
        self.n_epochs    = kwargs['n_epochs'] if 'n_epochs' in kwargs else 100
        self.lr          = kwargs['G_lr'] if 'G_lr' in kwargs else 1e-3
        self.batch_size  = kwargs['batch_size'] if 'batch_size' in kwargs else 16
        self.n_preds     = kwargs['n_preds'] if 'n_preds' in kwargs else 10
        self.verbose     = kwargs['verbose'] if 'verbose' in kwargs else 1

        self.save_loss   = kwargs['save_loss'] if 'save_loss' in kwargs else False
        self.mse_loss    = nn.MSELoss()
        self.kl_loss     = bnn.BKLLoss(reduction='mean', last_layer_only=False)
        self.kl_weight   = kwargs['kl_weight'] if 'kl_weight' in kwargs else 1e-2

        if self.save_loss:
            self.loss_hist = []
    
    def fit(self, X, y):

        if self.model is None:
            self.model = BayesianRegressor(n_inputs=X.shape[-1], n_outputs=y.shape[-1], hidden_size=self.hidden_size)
        
        self.opt = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        
        X = Tensor(X)
        y = Tensor(y)
        X_train = TensorDataset(X, y)

        fit_loop = range(self.n_epochs)
        if self.verbose > 1:
            fit_loop = tqdm(fit_loop, unit='epochs')
        
        for _ in fit_loop:
            for X_batch, y_batch in DataLoader(X_train, batch_size=self.batch_size, shuffle=True, drop_last=True):
                
                y_pred   = self.model(X_batch)
                mse_loss = self.mse_loss(y_pred, y_batch)
                kl_loss  = self.kl_loss(self.model)
                loss     = mse_loss + self.kl_weight * kl_loss

                if self.save_loss:
                    self.loss_hist.append(loss)

                self.opt.zero_grad()
                loss.backward()
                self.opt.step()
    
    def predict(self, X):

        f   = [self.model(Tensor(X)).detach().numpy() for _ in range(self.n_preds)]
        f   = np.concatenate(f, axis=-1)
        std = np.std(f, axis=-1)
        f   = np.mean(f, axis=-1)

        return f, std



import numpy as np
from ..base import FunctionV2


class Holder(FunctionV2):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.dim = 2
        self.domain = np.tile([-10, 10], reps=(2, 1))
        self.name = "Holder"
        self.glob_min = np.array([
            [8.05502, 9.66459], 
            [-8.05502, -9.66459], 
            [-8.05502, 9.66459], 
            [8.05502, -9.66459]
        ])
        self.f_min = 0.0
    
    def func(self, x: np.ndarray) -> np.ndarray:
        y = -np.abs(
            np.sin(x[..., 0]) * np.cos(x[..., 1]) * np.exp(np.abs(1 - np.sqrt(x[..., 0] ** 2 + x[..., 1] ** 2) / np.pi))
        ) + 19.2085
        return np.expand_dims(y, -1)



class Acquisition:
    def __init__(self):
        '''
        Base class for an acquisition function.

        Parameters
        ----------
        y : np.array of shape (n_samples * func.n_obs, 1)
            Target values of the black-box function.
        
        name : str
            Name of the acquisition function.
        
        surr : # TODO: add type
            Surrogate model to compute the acquisition function.
        '''
        super(Acquisition, self).__init__()

        self.y     = None
        self.name  = None
        self.surr  = None
        self.conds = None
    
    def __call__(self):
        '''
        Call of the acquisition function.

        Parameters
        ----------
        x : np.array of shape (n_samples * func.n_obs, func.dim)
            Candidate points.

        Returns
        -------
        Acquisition function value.
        '''
        pass


import numpy as np
from ..base import FunctionV2


class Levi(FunctionV2):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.dim = 2
        self.domain = np.tile([-4, 6], reps=(2, 1))
        self.name = "Lévi"
        self.glob_min = np.ones((1, 2))
        self.f_min = 0.0
    
    def func(self, x: np.ndarray) -> np.ndarray:
        y = np.sin(3 * np.pi * x[..., 0]) ** 2
        y += ((x[..., 0] - 1) ** 2) * (1 + np.sin(3 * np.pi * x[..., 1]) ** 2)
        y += ((x[..., 1] - 1) ** 2) * (1 + np.sin(2 * np.pi * x[..., 1]) ** 2)
        return np.expand_dims(y, -1)



import numpy as np
from tqdm import tqdm

from ..base import Optimiser
from ..utils import create_dir


class DifferentialEvolutionOptimizer(Optimiser):
    """
    Differential evolution black-box function optimizer

    Parameters
    ----------
    """

    def __init__(self, func, **kwargs):
        super(DifferentialEvolutionOptimizer).__init__()

        self.func         = func
        self.max_iter     = kwargs.get('max_iter', 1000)
        self.eps          = kwargs.get('eps', 1e-1)
        self.error_type   = kwargs.get('error_type', 'f')
        self.n_candidates = kwargs.get('n_candidates', 11 ** 2)
        self.olhs         = kwargs.get('olhs', True)
        self.lhs_seed     = kwargs.get('lhs_seed', None)
        self.verbose      = kwargs.get('verbose', 1)
        self.save_results = kwargs.get('save_results', True)
        self.candidates   = self.create_candidates()

        # Evolution parameters
        self.mutation_rate  = kwargs.get('mutation_rate', 1.0)
        self.crossover_rate = kwargs.get('crossover_rate', 0.5)

        # Experiments parameters
        self.errors = None
        self.res = None
        self.params = None

    def predict(self, X=None, y=None) -> np.ndarray:
        """
        Make one step of evolution and returns next best parameters.

        Parameters
        ----------
        X: Any, default = None
            Ignored
        y: Any, default = None
            Ignored
        
        Returns
        -------
        x_next: np.ndarray shape of [dim, ]
        """
        self.evolution_step()
        y_func = self.func(self.candidates).flatten() # shape of [n_candidates]
        return self.candidates[np.argmin(y_func)]
    
    def evolution_step(self):
        """
        Make evolution step: mutation, crossover, selecting.
        """
        X = self.candidates
        rng = np.random.default_rng()

        # Mutation
        Z = np.apply_along_axis(
            func1d = lambda x: rng.choice(X, size=3, replace=False),
            axis = -1,
            arr = X
        )

        R = self.mutation_rate
        coeffs = np.expand_dims([1.0, R, -R], axis=[-1, 0]) # shape of [1, 3, 1]
        Z = np.sum(Z * coeffs, axis=-2)

        # Crossover
        p = self.crossover_rate
        mask = rng.uniform(0, 1, Z.shape)
        mask = (mask < p).astype(bool)
        Z[mask] = X[mask]

        # Selection
        y_prev = self.func(X) # shape of [n_candidates, 1]
        y_next = self.func(Z) # shape of [n_candidates, 1]

        if y_prev.ndim == 1:
            y_prev = np.expand_dims(y_prev, -1)
        
        if y_next.ndim == 1:
            y_next = np.expand_dims(y_next, -1)
        
        self.candidates = np.where(
            (y_prev <= y_next), X, Z
        )

    def optimise(self, X=None, y=None, **kwargs):

        if X is None:
            X = self.create_candidates()
        self.candidates = X
        
        if self.verbose == 0:
            opt_loop = range(self.max_iter)
        else:
            opt_loop = tqdm(
                range(self.max_iter), 
                desc='Optimization loop started...', 
                leave=True, 
                position=0
            )

        for _ in opt_loop:
            self.predict()
            x_min, y_min, error = self.evaluate()

            # Log metrics            
            self.log_metrics(
                x_min = x_min,
                y_min = y_min,
                error = error
            )

            if error <= self.eps:
                print(f'Experiment finished successfully')
                break
        
        if error > self.eps:
            print('Experiment failed')
        
        if self.save_results:
            self._save()

    def evaluate(self) -> tuple[np.ndarray, np.ndarray, float]:
        x_cand = self.candidates
        y_pred = self.func(x_cand).flatten() # works only for 1-d output

        dim = self.func.dim
        x_glob_min = self.func.glob_min
        y_glob_min = self.func(self.func.glob_min.reshape(1, dim)).flatten()

        if self.error_type == 'x':
            error = np.min(
                np.linalg.norm(x_glob_min - x_cand, axis=-1),
                axis = -1
            )
        elif self.error_type == "f":
            error = np.min(
                np.linalg.norm(y_glob_min - y_pred, axis=-1),
                axis = -1
            )
        else:
            raise ValueError(f"Unsupported error type: {self.error_type}")
    
        x_min = x_cand[np.argmin(y_pred)]
        y_min = np.min(y_pred)

        return x_min, y_min, error
    
    def log_metrics(self, **kwargs) -> None:
        """
        Save optimization results
        """
        error = kwargs.pop('error')

        if self.errors is None:
            self.errors = []
        self.errors.append(error)

        x_min = kwargs.pop('x_min')
        y_min = kwargs.pop('y_min')

        if self.params is None:
            dim = self.func.dim
            self.params = np.zeros((0, dim))
            self.res = np.zeros((0, 1))
            
            x_last = None
            y_last = np.inf
        else:
            x_last = self.params[-1]
            y_last = self.res[-1, 0]


        if y_last <= y_min:
            x_best, y_best = x_last, y_last
        else:
            x_best, y_best = x_min, y_min
        
        self.params = np.concatenate((
            self.params, np.expand_dims(x_best, 0)
        ))
        self.res = np.concatenate((
            self.res, np.expand_dims([y_best], 0)
        ))



import time
import pickle
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm, ticker
from scipy.optimize import minimize
from joblib import Parallel, delayed

from .base import Optimiser
from .utils import create_dir


class SurrogateOptimiser(Optimiser):
    '''
    Surrogate based black-box optimiser.

    Parameter
    ----------
    func : Callable, waggon.functions.Function
        Black-box function to be optimised.
    
    surr : waggon.surrogates.Surrogate or waggon.surrogate.GenSurrogate
        Surrogate model for the black-box function.
    
    acqf : waggon.acquisitions.Acquisition
        Acquisition function defining the optimisation strategy.
    
    max_iter : int, default = 100
        Maximum number of optimisation loop iterations.
    
    eps : float, default = 1e-1
        Epsilon-solution criterion value.
    
    error_type : {'x', 'f'}, default = 'x'
        Optimisation error type - either in argument, 'x', or function, 'f'.
    
    num_opt : bool, default = False
        Whether the acquisition function is optimised numerically or not. If False,
        the search of the next best parameters is done via direct search.
    
    eq_cons : dict, default = None
        Equality-type constraints for numerical optimisation.
    
    ineq_cons : dict, default = None
        Ineqaulity-type constraints for numerical optimisation.

    fix_candidates : bool, default = False if num_opt else True
        Whether the candidate points should be fixed or not.
    
    n_candidates_ : int, default = 1 if num_opt else 101**2
        Number of candidates points.
    
    olhs : bool, default = True
        Whether orthogonal Latin hypercube sampling (LHS) is used for choosing candidate points.
        If False, simple LHS is used.

    lhs_seed : int, default = None
        Controls the randomness of candidates sampled via (orthogonal) LHS.
    
    verbose : int, default = 1
        Controls verbosity when fitting and predicting. By default only a progress bar over the
        optimisation loop is displayed.
    
    save_results : bool, default = True
        Whether results are saved or not.
    '''
    def __init__(self, func, surr, acqf, **kwargs):
        kwargs['func'] = func
        super(SurrogateOptimiser, self).__init__(**kwargs)
        
        self.func               = func
        self.surr               = surr
        self.acqf               = acqf
        
        self.n_candidates       = kwargs['n_candidates'] if 'n_candidates' in kwargs else 10201
        self.num_opt_start      = kwargs['num_opt_start'] if 'num_opt_start' in kwargs else 'grid'
        self.num_opt_disp       = kwargs['num_opt_disp'] if 'num_opt_disp' in kwargs else False
        self.num_opt_tol        = kwargs['num_opt_tol'] if 'num_opt_tol' in kwargs else 1e-6
        self.num_opt_candidates = kwargs['num_opt_candidates'] if 'num_opt_candidates' in kwargs else 128

        self.eq_cons            = kwargs['eq_cons'] if 'eq_cons' in kwargs else None
        self.ineq_cons          = kwargs['ineq_cons'] if 'ineq_cons' in kwargs else None

        self.jitter             = kwargs['jitter'] if 'jitter' in kwargs else 1e0
        self.parallel           = kwargs['parallel'] if 'parallel' in kwargs else 0
        
        if type(self.surr) == list:
            for s in self.surr:
                s.verbose = self.verbose
        else:
            self.surr.verbose   = self.verbose
    
    
    def run_lbfgsb(self, x0):
        opt_res = minimize(method='L-BFGS-B', fun=self.acqf, x0=x0, bounds=self.func.domain, tol=self.num_opt_tol, options={'disp': self.num_opt_disp})
        return opt_res.fun, opt_res.x
    

    def numerical_search(self, x0=None):
        '''
        Numerical optimisation of the acquisition function.

        Returns
        -------
        best_x : np.array of shape (func.dim,)
            Predicted optimum of the acquisition function.
        '''
        
        if (x0 is None) and (self.num_opt_start != 'grid'):
            candidates = self.create_candidates()
        elif self.num_opt_start == 'random':
            candidates = np.array(self.num_opt_candidates * [x0])
            candidates += np.random.normal(0, self.eps, candidates.shape)
        elif self.num_opt_start == 'grid':
            inter_conds = self.create_candidates(N=self.n_candidates)
            if self.num_opt_candidates < self.n_candidates:
                ei = self.acqf(inter_conds)
                try:
                    ids = np.argsort(ei, axis=0)[:self.num_opt_candidates].reshape(-1, 1)
                    candidates = np.take_along_axis(inter_conds, ids, axis=0)
                except ValueError:
                    ei = ei.squeeze()
                    ids = np.argsort(ei, axis=0)[:self.num_opt_candidates].reshape(-1, 1)
                    candidates = np.take_along_axis(inter_conds, ids, axis=0)
            else:
                candidates = inter_conds
        
        if self.parallel in [0, 1]:

            for i, x0 in enumerate(candidates):
                
                if (self.eq_cons is None) and (self.ineq_cons is None):
                    opt_res = minimize(method='L-BFGS-B', fun=self.acqf, x0=x0, bounds=self.func.domain, tol=self.num_opt_tol, options={'disp': self.num_opt_disp})
                else:
                    opt_res = minimize(method='SLSQP', fun=self.acqf, x0=x0, bounds=self.func.domain, constraints=[self.eq_cons, self.ineq_cons])
                
                if i == 0:
                    best_x = opt_res.x
                    best_acqf = opt_res.fun
                else:
                    if opt_res.fun < best_acqf:
                        best_acqf = opt_res.fun
                        best_x = opt_res.x
            
            return best_x
        
        else:

            r = Parallel(n_jobs=self.parallel, prefer="threads")(delayed(self.run_lbfgsb)(x0) for x0 in candidates)
            f, x = zip(*r)
            f, x = np.array(f), np.array(x)

            return x[np.argmin(f)]
    

    def predict(self, X, y):
        '''
        Predicts the next best set of parameter values by optimising the acquisition function.
        
        Parameters
        ----------
        X : np.array of shape (n_samples * func.n_obs, func.dim)
            Training input points.

        y : np.array of shape (n_samples * func.n_obs, func.dim)
            Target values of the black-box function.
        
        n_pred : int, defult = 1
            Number of predictions to make.
        
        Returns
        -------
        next_x : np.array of shape (func.dim, n_pred)
        '''
        
        next_x = np.nan

        while not np.any(np.sum(X - next_x, axis=-1) < 1e-6):

            self.surr.fit(X, y)
            
            self.acqf.y = y.reshape(y.shape[0]//self.func.n_obs, self.func.n_obs)
            self.acqf.conds = X[::self.func.n_obs]
            self.acqf.surr = self.surr

            x0 = None
            if self.num_opt_start == 'fmin':
                x0 = X[np.argmin(y)]
            
            next_x = self.numerical_search(x0=x0)

        return np.array([next_x])
    
    
    def _save(self, base_dir='test_results'):
        res_path = create_dir(self.func, self.acqf.name, self.surr.name, base_dir=base_dir)

        res = {'X': self.params,
               'y': self.res,
               'err': self.errors}

        with open(f'{res_path}/{time.strftime("%d_%m_%H_%M_%S")}.pkl', 'wb') as f:
            pickle.dump(res, f)
    

    def plot_iteration_results(self, X, next_x):
        '''
        For surrogate optimiser only.
        '''
        if self.func.dim == 1:
            inter_conds = np.linspace(self.func.domain[:, 0], self.func.domain[:, 1], 121)
        else:
            inter_conds = self.create_candidates(N=10201)
        
        # transform = lambda x: np.exp(x) if self.func.log_transform else lambda x: x

        mu, _ = self.surr.predict(inter_conds)
        mu = mu
        y_true = self.func(inter_conds)
        
        if self.func.dim == 1:
            plt.plot(inter_conds, mu, label=f'Pred: {np.mean((y_true - mu)**2):.2f}', c='cornflowerblue')
            plt.plot(inter_conds, y_true, label='True', c='orange')
            # plt.scatter(X, y, c='black')
            plt.legend()
        else:
            
            y_true, mu = y_true.reshape(101, 101), mu.reshape(101, 101)
            mse = (y_true - mu)**2

            ei = self.acqf(inter_conds).reshape(101, 101)
            x_pred = inter_conds[np.argmin(ei)]
            
            def single_2d_plot(axis, f, title):
                plt.subplot(axis)
                colormap = plt.contourf(f, locator=ticker.LinearLocator(), extent=self.func.domain.flatten(), vmin=np.min(f), vmax=np.max(f))
                plt.scatter(X[:, 0], X[:, 1], color='black')
                for gb in self.func.glob_min:
                    plt.scatter(gb[0], gb[1], color='red', marker='*')
                plt.scatter(x_pred[0], x_pred[1], color='cyan', marker='*')
                plt.scatter(next_x[0], next_x[1], color='magenta', marker='*')
                plt.title(title)
                plt.colorbar(colormap)
            
            plt.figure(figsize=(24, 6))

            single_2d_plot(141, y_true, 'True Function')
            single_2d_plot(142, mu, 'Estimated Function')
            single_2d_plot(143, mse, 'MSE')
            single_2d_plot(144, ei, 'EI')
            
            plt.tight_layout()
        
        plt.show()

